/*
* Author: Chris Campbell - www.iforce2d.net
*
* Copyright (c) 2006-2011 Erin Catto http://www.box2d.org
*
* This software is provided 'as-is', without any express or implied
* warranty.  In no event will the authors be held liable for any damages
* arising from the use of this software.
* Permission is granted to anyone to use this software for any purpose,
* including commercial applications, and to alter it and redistribute it
* freely, subject to the following restrictions:
* 1. The origin of this software must not be misrepresented; you must not
* claim that you wrote the original software. If you use this software
* in a product, an acknowledgment in the product documentation would be
* appreciated but is not required.
* 2. Altered source versions must be plainly marked as such, and must not be
* misrepresented as being the original software.
* 3. This notice may not be removed or altered from any source distribution.
*/

#ifndef IFORCE2D_ENGINE_H
#define IFORCE2D_ENGINE_H

#include <vector>
#include <set>

#include "../Framework/b2dJson.h"

#ifndef DEGTORAD
#define DEGTORAD 0.0174532925199432957f
#define RADTODEG 57.295779513082320876f
#endif

#define NUM_GRAPHPOINTS 120

#include "iforce2d_Buoyancy_functions.h"

class PowerGraph {
public:

    float torquePoints[NUM_GRAPHPOINTS];
    int pointCounts[NUM_GRAPHPOINTS];
    PowerGraph() {
        memset(pointCounts, 0, sizeof(pointCounts));
    }

    void addMeasurement(float rpm, float torque) {
        int i = (int)(rpm + 0.5f);
        if ( i < 0 || i >= NUM_GRAPHPOINTS )
            return;
        float uptake = pointCounts[i] < 2 ? 1 : (1/(float)(pointCounts[i]-1));
        torquePoints[i] = (1-uptake) * torquePoints[i] + uptake * torque;
        pointCounts[i]++;
    }

    void draw( b2Vec2 offset, b2Vec2 scale ) {
        glEnable(GL_BLEND);
        glEnable(GL_POINT_SMOOTH);
        glPushMatrix();
        glTranslatef( offset.x, offset.y, 0 );
        glScalef( scale.x, scale.y, 1 );

        //frame
        glColor3f(0.75f, 0.75f, 0.75f);
        glBegin(GL_LINE_LOOP);
        glVertex2f( 0,  0);
        glVertex2f( 1,  0);
        glVertex2f( 1,  1);
        glVertex2f( 0,  1);
        glEnd();

        glPointSize(8);
        glBegin(GL_POINTS);
        for (int i = 0; i < (NUM_GRAPHPOINTS-1); i++) {
            if ( pointCounts[i] < 5 )
                continue;
            float alpha = pointCounts[i] * 0.001f;
            alpha = b2Max(0.1f, alpha);
            glColor4f(1,0,0,alpha);
            glVertex2f( i * (1/(float)NUM_GRAPHPOINTS), 0.00002f * torquePoints[i] );
            glColor4f(0,1,0,alpha);
            glVertex2f( i * (1/(float)NUM_GRAPHPOINTS), 0.00002f * torquePoints[i] * i * 0.025f );
        }
        glEnd();

        glBegin(GL_POINTS);
        for (int i = 15; i < (NUM_GRAPHPOINTS-1); i++) {
            int range = 8;
            float gauss = 0;
            int count = 0;
            float totalWeight = 0;
            for (int k = i-range; k <= i+range; k++) {
                if ( k < 0 || k >= NUM_GRAPHPOINTS || pointCounts[k] < 20 )
                    continue;
                int dist = abs(i-k);
                float weight = (range - dist)/(float)range;
                gauss += weight * weight * pointCounts[k] * torquePoints[k];
                count++;
                totalWeight += weight * weight * pointCounts[k];
            }
            if ( count < range )
                continue;
            gauss /= totalWeight;
            //gauss /= (float)count;
            glColor3f(1,0.5f,1);
            glVertex2f( i * (1/(float)NUM_GRAPHPOINTS), 0.00002f * gauss );
            glColor3f(1,1,0.5f);
            glVertex2f( i * (1/(float)NUM_GRAPHPOINTS), 0.00002f * gauss * i * 0.025f );
        }
        glEnd();

        glPopMatrix();
    }

    void drawCurrentRpmMarker( b2Vec2 offset, b2Vec2 scale, float rpm) {
        int i = (int)(rpm + 0.5f);
        if ( i < 0 || i >= NUM_GRAPHPOINTS )
            return;
        glColor4f(1,1,0,0.25f);

        glPushMatrix();
        glTranslatef( offset.x, offset.y, 0 );
        glScalef( scale.x, scale.y, 1 );

        glBegin(GL_LINES);
        glVertex2f( i * (1/(float)NUM_GRAPHPOINTS), 0 );
        glVertex2f( i * (1/(float)NUM_GRAPHPOINTS), 1 );
        glEnd();

        glPopMatrix();
    }
};

int g_engineDesign = 0;
bool g_throttleOn = true;
int g_burnSteps = 10;

PowerGraph g_powerGraphs[3];

class FuelParticleData {
public:
    float m_energy;//if non-zero this has not been burned yet
    int m_burnStepsRemaining;//if non-zero this particle can set another particle alight

    FuelParticleData() { m_energy = 1; m_burnStepsRemaining = 0; }
};

const char* designs[] = {
    "fuelDemo",
    "demoCylinder",
    "singleCylinder",
    "dualCylinder",
    "quadCylinder",
    "boatEngine"
};

class iforce2d_combustionEngine : public Test
{
public:
    iforce2d_combustionEngine()
    {
        //replace the testbed's default world with the one saved in json file
        b2dJson json;
        string filename = string(designs[g_engineDesign]) + ".json";
        b2World* world = json.readFromFile(filename.c_str());
        if ( world ) {
            delete m_world;
            m_world = world;

            m_world->SetContactListener(this);
            m_world->SetDebugDraw(&m_debugDraw);

            b2BodyDef bodyDef;
            m_groundBody = m_world->CreateBody(&bodyDef);

            //find important things
            m_engineBody = json.getBodyByName("engine");
            m_crankshaftJoint = (b2RevoluteJoint*)json.getJointByName("crankshaft");
            m_flywheelBody = json.getBodyByName("flywheel");
            json.getFixturesByName("catcher", m_catchers);

            vector<b2Fixture*> carbFixtures;
            json.getFixturesByName("carb", carbFixtures);
            for (int i = 0; i < carbFixtures.size(); i++)
                m_carbPositions.push_back(((b2CircleShape*)carbFixtures[i]->GetShape())->m_p);

            for (int i = 0; i < 4; i++) {
                char buf[16];
                sprintf(buf, "spark%d", i);
                b2Fixture* f = json.getFixtureByName(buf);
                if ( !f )
                    break;
                if ( f ) {
                    m_sparkPositions.push_back( ((b2CircleShape*)f->GetShape())->m_p );
                    m_sparkOn.push_back(false);
                }
            }

            for (int i = 0; i < 4; i++) {
                char buf[16];
                sprintf(buf, "ignition%d", i);
                vector<b2Fixture*> ignitionPair;
                if ( 2 == json.getFixturesByName(buf, ignitionPair) )
                    m_ignitionFixturePairs.push_back( make_pair(ignitionPair[0], ignitionPair[1]) );
            }

            for (int i = 0; i < 8; i++) {
                char buf[16];
                sprintf(buf, "valve%d", i);
                b2Fixture* f = json.getFixtureByName(buf);
                if ( !f )
                    break;
                m_valveFixtures.push_back(f);
                sprintf(buf, "blocker%d", i);
                m_blockerFixtures.push_back( json.getFixtureByName(buf) );
                m_valvesOpen.push_back(false);
            }

            vector<b2Joint*> camshaftJoints;
            json.getJointsByName("camshaft", camshaftJoints);
            for (int i = 0; i < (int)camshaftJoints.size(); i++) {
                b2GearJointDef jd;
                jd.joint1 = m_crankshaftJoint;
                jd.joint2 = (b2RevoluteJoint*)camshaftJoints[i];
                jd.bodyA = jd.joint1->GetBodyB();
                jd.bodyB = jd.joint2->GetBodyB();
                jd.ratio = 2;
                m_world->CreateJoint(&jd);
            }

            m_torqueGuageFrictionJoint = NULL;
            m_torqueGuageMotorJoint = NULL;

            m_torqueGuageBody = json.getBodyByName("torqueBar");
            if ( m_torqueGuageBody )
            {
                b2MotorJointDef mjd;
                mjd.bodyA = m_flywheelBody;
                mjd.bodyB = m_torqueGuageBody;
                mjd.maxForce= 0;
                mjd.maxTorque = FLT_MAX;//needs to be enough to stop engine
                m_torqueGuageMotorJoint = (b2MotorJoint*)m_world->CreateJoint(&mjd);

                b2FrictionJointDef fjd;
                fjd.bodyA = m_engineBody;
                fjd.bodyB = m_torqueGuageBody;
                fjd.maxForce = 0;
                fjd.maxTorque = 0;
                m_torqueGuageFrictionJoint = (b2FrictionJoint*)m_world->CreateJoint(&fjd);
            }

            //m_torqueSpringJoint = (b2DistanceJoint*)json.getJointByName("torqueSpring");

            m_waterFixture = json.getFixtureByName("water");
            m_boathullFixture = json.getFixtureByName("boathull");

            //m_brakePistonJoint = (b2PrismaticJoint*)json.getJointByName("brakepiston");
            //m_brakePadJoint = (b2PrismaticJoint*)json.getJointByName("brakepad");

            m_runningDyno = false;

        }

        m_rotationSinceLastFuel = 0;
        m_fuelRate = 0;
        m_currentFriction = 0;
        m_currentTorque = 0;
        m_currentRpm = 0;
        m_torqueForGraph = 0;
        memset(m_torqueMeasurements, 0, sizeof(m_torqueMeasurements));

        m_spinCrank = false;

        glEnable(GL_POINT_SMOOTH);
    }

    void Keyboard(unsigned char key)
    {
        switch (key) {
        case 'a': g_engineDesign = (g_engineDesign+1) % (sizeof(designs)/sizeof(char*)); break;
        case 's': g_throttleOn = !g_throttleOn; break;
        //case 'd': g_burnSteps--; break;
        //case 'e': g_burnSteps++; break;
        //case 'e': if ( m_brakePistonJoint ) m_brakePistonJoint->SetMaxMotorForce( m_brakePistonJoint->GetMaxMotorForce() + 1000 ); break;
        //case 'd': if ( m_brakePistonJoint ) m_brakePistonJoint->SetMaxMotorForce( b2Max(0.0f, m_brakePistonJoint->GetMaxMotorForce() - 1000 )); break;
        case 'e': if ( m_torqueGuageFrictionJoint ) m_torqueGuageFrictionJoint->SetMaxTorque( m_torqueGuageFrictionJoint->GetMaxTorque() + 1000 ); break;
        case 'd': if ( m_torqueGuageFrictionJoint ) m_torqueGuageFrictionJoint->SetMaxTorque( b2Max(0.0f, m_torqueGuageFrictionJoint->GetMaxTorque() - 1000 )); break;
        case 'c': m_spinCrank = true; break;
        case 'g': m_runningDyno = !m_runningDyno; break;
        default: Test::Keyboard(key);
        }
        if ( g_burnSteps < 0 )
            g_burnSteps = 0;
    }

    void KeyboardUp(unsigned char key)
    {
        switch (key)
        {
        case 'c': m_spinCrank = false; break;
        default: Test::Keyboard(key);
        }
    }

    void igniteParticle(b2Body* body)
    {
        FuelParticleData* fpd = (FuelParticleData*)body->GetUserData();
        fpd->m_energy = 0;
        fpd->m_burnStepsRemaining = g_burnSteps;
        body->GetFixtureList()->SetDensity(750);
        body->ResetMassData();
        body->SetLinearDamping(0.01);
        b2Vec2 dir = body->GetLinearVelocity();
        dir.Normalize();
        body->SetLinearVelocity( 200 * dir );
    }

    void BeginContact(b2Contact* contact)
    {
        b2Fixture* fA = contact->GetFixtureA();
        b2Fixture* fB = contact->GetFixtureB();

        for (int i = 0; i < m_valveFixtures.size(); i++) {
            //check for valve touching sensor, means the valve is open
            if ( fA == m_valveFixtures[i] ) {
                if ( fB->IsSensor() )
                    m_valvesOpen[i] = true;
            }
            if ( fB == m_valveFixtures[i] ) {
                if ( fA->IsSensor() )
                    m_valvesOpen[i] = true;
            }

            //when intake valve is open, contacts between the blocker and fuel should be disabled
            if ( m_valvesOpen[i] ) {
                if ( g_engineDesign == 1 ) {//demo cylinder has no blocker fixtures
                    if ( (fA == m_valveFixtures[i] && fB->GetShape()->GetType() == b2Shape::e_circle) ||
                         (fB == m_valveFixtures[i] && fA->GetShape()->GetType() == b2Shape::e_circle) )
                        contact->SetEnabled(false);
                }
                else {
                    if ( (fA == m_blockerFixtures[i] && fB->GetShape()->GetType() == b2Shape::e_circle) ||
                         (fB == m_blockerFixtures[i] && fA->GetShape()->GetType() == b2Shape::e_circle) )
                        contact->SetEnabled(false);
                }
            }
        }


        //catch particles at the end of the exhaust and mark them for recycling
        for (int i = 0; i < m_catchers.size(); i++) {
            if ( fA == m_catchers[i] && fB->GetFilterData().categoryBits == 2 )
                m_particlesToRecycle.push_back( fB->GetBody() );
            if ( fB == m_catchers[i] && fA->GetFilterData().categoryBits == 2 )
                m_particlesToRecycle.push_back( fA->GetBody() );
        }


        //check for burning particles hitting un-burned fuel
        if ( fA->GetBody()->GetUserData() && fB->GetBody()->GetUserData() ) {
            FuelParticleData* fpdA = (FuelParticleData*)fA->GetBody()->GetUserData();
            FuelParticleData* fpdB = (FuelParticleData*)fB->GetBody()->GetUserData();
            if ( fpdA->m_energy == 0 && fpdB->m_energy == 0 ) {
            //if ( (fpdA->m_energy == 0 && fpdA->m_burnStepsRemaining < 1) || fpdB->m_energy == 0 && fpdB->m_burnStepsRemaining < 1 ) {
                contact->SetEnabled(false);
                return;
            }
            if ( fpdA->m_burnStepsRemaining > 0 && fpdB->m_energy > 0 /*&& m_engineBody->GetLocalPoint( fB->GetBody()->GetPosition() ).y < 8*/ ) {
                igniteParticle( fB->GetBody() );
            }
            else if ( fpdB->m_burnStepsRemaining > 0 && fpdA->m_energy > 0 /*&& m_engineBody->GetLocalPoint( fA->GetBody()->GetPosition() ).y < 8*/ ) {
                igniteParticle( fA->GetBody() );
            }
        }

        //set flag to fire spark after time step
        for (int i = 0; i < m_ignitionFixturePairs.size(); i++) {
            b2Fixture* f0 = m_ignitionFixturePairs[i].first;
            b2Fixture* f1 = m_ignitionFixturePairs[i].second;
            if ( (fA == f0 && fB == f1) ||
                 (fA == f1 && fB == f0))
                m_sparkOn[i] = true;
        }

        //buoyancy
        if ( fA == m_waterFixture )
            m_buoyancyFixturePairs.insert( make_pair(fA, fB) );
        else if ( fB == m_waterFixture )
            m_buoyancyFixturePairs.insert( make_pair(fB, fA) );
    }

    /*void reenableIntakeBlocker()
    {
        //intake valve has closed, re-enable the blocker fixure (helps prevent
        //misfiring due to flame traveling through the valve via a fuel particle
        //that was inside the valve fixture
        b2ContactEdge* edge = m_engineBody->GetContactList();
        while (edge) {
            edge->contact->SetEnabled(true);
            edge = edge->next;
        }
    }*/

    void EndContact(b2Contact* contact)
    {
        b2Fixture* fA = contact->GetFixtureA();
        b2Fixture* fB = contact->GetFixtureB();
        for (int i = 0; i < m_valveFixtures.size(); i++) {
            if ( fA == m_valveFixtures[i] ) {
                if ( fB->IsSensor() ) {
                    m_valvesOpen[i] = false;
                    //if ( i == 0 )
                    //    reenableIntakeBlocker();
                }
            }
            if ( fB == m_valveFixtures[i] ) {
                if ( fA->IsSensor() ) {
                    m_valvesOpen[i] = false;
                    //if ( i == 0 )
                    //    reenableIntakeBlocker();
                }
            }
        }

        //buoyancy
        if ( fA == m_waterFixture && fB == m_boathullFixture )
            m_buoyancyFixturePairs.erase( make_pair(fA, fB) );
        else if ( fB == m_waterFixture && fA == m_boathullFixture )
            m_buoyancyFixturePairs.erase( make_pair(fB, fA) );
    }

    b2Body* getQueuedParticle()
    {
        b2Vec2 queuedPos = b2Vec2(-20, -20) + m_queuedParticles.size() * b2Vec2(0.3, 0);

        if ( m_queuedParticles.empty() ) {
            b2BodyDef bd;
            bd.type = b2_dynamicBody;
            bd.bullet = true;
            bd.fixedRotation = true;
            bd.linearDamping = 0.1;
            bd.gravityScale = 0;
            bd.position = queuedPos;
            b2Body* body = m_world->CreateBody(&bd);

            body->SetUserData( new FuelParticleData() );

            b2CircleShape shape;
            shape.m_radius = 0.15f;

            b2FixtureDef fd;
            fd.shape = &shape;
            fd.density = 1;
            fd.friction = 0;
            fd.restitution = 1;
            fd.filter.categoryBits = 2;
            body->CreateFixture(&fd);

            m_queuedParticles.push_back(body);
        }

        b2Body* body = m_queuedParticles.back();
        body->GetFixtureList()->SetDensity(1);
        body->ResetMassData();
        m_queuedParticles.pop_back();
        return body;
    }

    void addFuel()
    {
        if ( m_fuelParticles.size() >= 40 * m_sparkPositions.size() )
            return;

        for (int i = 0; i < m_carbPositions.size(); i++) {
            b2Vec2 pos = m_engineBody->GetWorldPoint( m_carbPositions[i] );
            b2Vec2 vel = 40 * b2Vec2( -0.5f + rnd_1(), -0.5f + rnd_1() );//m_engineBody->GetWorldVector( b2Vec2(-10, 9 + rnd_1() * 2 ) );

            b2Body* body = getQueuedParticle();
            body->SetTransform( pos, 0 );
            body->SetLinearVelocity( vel );
            m_fuelParticles.insert(body);

            FuelParticleData* fpd = (FuelParticleData*)body->GetUserData();
            fpd->m_energy = 1;
            fpd->m_burnStepsRemaining = 0;
        }
    }

    void addSpark(int whichPos)
    {
        int numSparks = g_engineDesign == 0 ? 1 : 3;
        for (int i = 0 ; i < numSparks; i++) {
            b2Vec2 sparkPos = m_sparkPositions[whichPos];
            b2Vec2 pos = m_engineBody->GetWorldPoint( sparkPos );
            b2Vec2 vel = 10 * (m_crankshaftJoint->GetAnchorA() + b2Vec2(-5+i*5,0) - pos);

            b2Body* body = getQueuedParticle();
            body->SetTransform( pos, 0 );
            body->SetLinearVelocity( vel );
            m_fuelParticles.insert(body);

            FuelParticleData* fpd = (FuelParticleData*)body->GetUserData();
            fpd->m_energy = 0;
            fpd->m_burnStepsRemaining = g_burnSteps;
        }
    }

    void updateBuoyancyForces()
    {
        //go through all buoyancy fixture pairs and apply necessary forces
        std::set<fixturePair>::iterator it = m_buoyancyFixturePairs.begin();
        std::set<fixturePair>::iterator end = m_buoyancyFixturePairs.end();
        while (it != end) {

            //fixtureA is the fluid
            b2Fixture* fixtureA = it->first;
            b2Fixture* fixtureB = it->second;

            float density = fixtureA->GetDensity();

            std::vector<b2Vec2> intersectionPoints;
            if ( findIntersectionOfFixtures(fixtureA, fixtureB, intersectionPoints) ) {

                //find centroid
                float area = 0;
                b2Vec2 centroid = ComputeCentroid( intersectionPoints, area);

                //apply buoyancy force
                float displacedMass = fixtureA->GetDensity() * area;
                b2Vec2 gravity( 0, -10 );
                b2Vec2 buoyancyForce = displacedMass * -gravity;
                fixtureB->GetBody()->ApplyForce( buoyancyForce, centroid, true );

                //apply complex drag
                float dragMod = 0.25f;//adjust as desired
                float liftMod = 0.25f;//adjust as desired
                float maxDrag = 2000;//adjust as desired
                float maxLift = 500;//adjust as desired
                for (int i = 0; i < intersectionPoints.size(); i++) {
                    b2Vec2 v0 = intersectionPoints[i];
                    b2Vec2 v1 = intersectionPoints[(i+1)%intersectionPoints.size()];
                    b2Vec2 midPoint = 0.5f * (v0+v1);

                    //find relative velocity between object and fluid at edge midpoint
                    b2Vec2 velDir = fixtureB->GetBody()->GetLinearVelocityFromWorldPoint( midPoint ) -
                                    fixtureA->GetBody()->GetLinearVelocityFromWorldPoint( midPoint );
                    float vel = velDir.Normalize();

                    b2Vec2 edge = v1 - v0;
                    float edgeLength = edge.Normalize();
                    b2Vec2 normal = b2Cross(-1,edge);
                    float dragDot = b2Dot(normal, velDir);
                    if ( dragDot < 0 )
                        continue;//normal points backwards - this is not a leading edge

                    //apply drag
                    float dragMag = dragDot * dragMod * edgeLength * density * vel * vel;
                    dragMag = b2Min( dragMag, maxDrag );
                    b2Vec2 dragForce = dragMag * -velDir;
                    fixtureB->GetBody()->ApplyForce( dragForce, midPoint, true );

                    //apply lift
                    float liftDot = b2Dot(edge, velDir);
                    float liftMag =  dragDot * liftDot * liftMod * edgeLength * density * vel * vel;
                    liftMag = b2Min( liftMag, maxLift );
                    b2Vec2 liftDir = b2Cross(1,velDir);
                    b2Vec2 liftForce = liftMag * liftDir;
                    fixtureB->GetBody()->ApplyForce( liftForce, midPoint, true );
                }

                //draw debug info
                glColor3f(0,1,1);
                glLineWidth(2);
                glBegin(GL_LINE_LOOP);
                for (int i = 0; i < intersectionPoints.size(); i++)
                    glVertex2f( intersectionPoints[i].x, intersectionPoints[i].y );
                glEnd();
                glLineWidth(1);
            }

            ++it;
        }
    }

    void Step(Settings* settings)
    {
        bool wasSingleStep = settings->singleStep;

        Test::Step(settings);

        glEnable(GL_BLEND);
        glEnable(GL_POINT_SMOOTH);

        if ( !settings->pause || wasSingleStep ) {

            for (int i = 0; i < m_particlesToRecycle.size(); i++) {
                b2Vec2 queuedPos = b2Vec2(-20, -20) + m_queuedParticles.size() * b2Vec2(0.3, 0);
                m_particlesToRecycle[i]->SetTransform( queuedPos, 0 );
                m_particlesToRecycle[i]->SetLinearVelocity( b2Vec2(0,0) );
                m_queuedParticles.push_back( m_particlesToRecycle[i] );
                m_fuelParticles.erase( m_particlesToRecycle[i] );
            }
            m_particlesToRecycle.clear();

            if ( g_throttleOn ) {
                m_rotationSinceLastFuel += -0.5 * m_crankshaftJoint->GetJointSpeed();
                if ( m_rotationSinceLastFuel > 10 ) {
                    addFuel();
                    m_rotationSinceLastFuel = 0;
                }
            }

            set<b2Body*>::iterator it = m_fuelParticles.begin();
            set<b2Body*>::iterator end = m_fuelParticles.end();
            while (it != end) {
                FuelParticleData* fpd = (FuelParticleData*)(*it)->GetUserData();
                if ( fpd->m_burnStepsRemaining > 0 )
                    fpd->m_burnStepsRemaining--;
                if ( fpd->m_burnStepsRemaining == 0 ) {
                    (*it)->GetFixtureList()->SetDensity(1);
                    (*it)->ResetMassData();
                }
                ++it;
            }

            if ( m_crankshaftJoint ) {
                float rpmUptake = 1/300.0f;//average last five seconds
                float rpmNow = -m_crankshaftJoint->GetJointSpeed() * RADTODEG / 360.f * 60;
                m_currentRpm = (1-rpmUptake) * m_currentRpm + rpmUptake * rpmNow;
            }

            if ( m_torqueGuageMotorJoint ) {
                float uptake = 1/300.0f;//average last five seconds

                float frictionNow = m_torqueGuageFrictionJoint->GetMaxTorque();
                m_currentFriction = (1-uptake) * m_currentFriction + uptake * frictionNow;

                float torqueNow = -m_torqueGuageMotorJoint->GetReactionTorque(settings->hz);
                m_currentTorque = (1-uptake) * m_currentTorque + uptake * torqueNow;

                //motor joint should just try to keep relative rotation to zero, rather than move to some specific target angle
                m_torqueGuageMotorJoint->SetAngularOffset( m_torqueGuageMotorJoint->GetAngularOffset() );

                for (int i = 0; i < 599; i++)
                    m_torqueMeasurements[i] = m_torqueMeasurements[i+1];

                uptake = 1/10.0f;//average last ten steps
                m_torqueForGraph = (1-uptake) * m_torqueForGraph + uptake * torqueNow;
                m_torqueMeasurements[599] = m_torqueForGraph;

                if ( g_engineDesign > 1 && g_engineDesign < 5 && m_currentFriction >= 10 )
                    g_powerGraphs[g_engineDesign-2].addMeasurement( m_currentRpm, m_currentTorque * 0.1f );
            }

            updateBuoyancyForces();

            if ( m_spinCrank )
                m_flywheelBody->ApplyAngularImpulse( -3000, true );

            if ( m_runningDyno ) {
                if ( m_currentRpm > 1 )
                    m_torqueGuageFrictionJoint->SetMaxTorque( m_torqueGuageFrictionJoint->GetMaxTorque() + 10 );
                else
                    m_runningDyno = false;
            }

        }

        for (int i = 0; i < m_sparkPositions.size(); i++) {
            if ( m_sparkOn[i] ) {
                addSpark(i);
                m_sparkOn[i] = false;
            }
        }

        //fuel particle colors
        glPointSize(8);
        glBegin(GL_POINTS);
        set<b2Body*>::iterator it = m_fuelParticles.begin();
        set<b2Body*>::iterator end = m_fuelParticles.end();
        while (it != end) {
            FuelParticleData* fpd = (FuelParticleData*)(*it)->GetUserData();
            if ( fpd->m_energy > 0 )
                glColor3f( 0, 0.75, 0.75 );
            else
                glColor3f( 0.75, fpd->m_burnStepsRemaining / (float)g_burnSteps, 0 );
            b2Vec2 pos = (*it)->GetPosition();
            glVertex2fv( (GLfloat*)&pos );
            ++it;
        }
        glEnd();
        glPointSize(1);


        //current torque line
        glPointSize(2);
        glColor3f(1,1,0);
        glBegin(GL_LINE_LOOP);
        glVertex2f( -50, 60 );
        for (int i = 0; i < 600; i++)
            glVertex2f( -50 + i*100/600.0, 60 + 0.00002f * m_torqueMeasurements[i] );
        glVertex2f( 50, 60 );
        glEnd();

        //power graphs
        if ( g_engineDesign > 1 && g_engineDesign < 5 ) {
            for (int i = 0; i < 3; i++)
                g_powerGraphs[i].draw( b2Vec2(-35+i*25, 75), b2Vec2(20, 20) );
            g_powerGraphs[g_engineDesign-2].drawCurrentRpmMarker( b2Vec2(-35+(g_engineDesign-2)*25, 75), b2Vec2(20, 20), m_currentRpm);
        }

        m_debugDraw.DrawString(5, m_textLine, "Press a to change engine design (change will take effect after restart: %s)", designs[g_engineDesign]);
        m_textLine += 15;

        m_debugDraw.DrawString(5, m_textLine, "Press s to toggle fuel: (currently %s)", g_throttleOn?"on":"off");
        m_textLine += 15;

        m_debugDraw.DrawString(5, m_textLine, "Hold c to crank engine: (currently %s)", m_spinCrank?"on":"off");
        m_textLine += 15;

        m_debugDraw.DrawString(5, m_textLine, "Press g to toggle dyno run: (currently %s)", m_runningDyno?"on":"off");
        m_textLine += 15;

        //m_debugDraw.DrawString(5, m_textLine, "Press d/e to change burn time: (currently %d time steps)", g_burnSteps);
        //m_textLine += 15;

        if ( m_torqueGuageFrictionJoint ) {
            m_debugDraw.DrawString(5, m_textLine, "Press d/e to change load: (currently %.3f)", m_torqueGuageFrictionJoint->GetMaxTorque() * 0.001f );
            m_textLine += 15;
        }

        m_debugDraw.DrawString(5, m_textLine, "Crankshaft rotation: %.1f rpm", m_currentRpm);
        m_textLine += 15;

        //m_debugDraw.DrawString(5, m_textLine, "Particles: %d", m_fuelParticles.size());
        //m_textLine += 15;

        m_debugDraw.DrawString(5, m_textLine, "Torque: %.1f", m_currentTorque *0.0001f);
        m_textLine += 15;

        /*if ( m_torqueGuageMotorJoint ) {

            float avgTorque = 0;
            for (int i = 0; i < 600; i++) {
                float t = m_torqueMeasurements[i];
                if ( t > 0 )
                    avgTorque += t;
            }
            avgTorque /= 600.0f;

            m_debugDraw.DrawString(5, m_textLine, "Torque: %.1f", avgTorque );
            m_textLine += 15;
        }*/

        /*m_debugDraw.DrawString(5, m_textLine, "Intake open: %s", m_valvesOpen[0]?"true":"false");
        m_textLine += 15;

        m_debugDraw.DrawString(5, m_textLine, "Exhaust open: %s", m_valvesOpen[1]?"true":"false");
        m_textLine += 15;*/
    }

    static Test* Create()
    {
        return new iforce2d_combustionEngine;
    }

    b2Body* m_engineBody;
    b2RevoluteJoint* m_crankshaftJoint;
    b2Body* m_flywheelBody;
    vector<b2Vec2> m_carbPositions;//local
    vector<b2Vec2> m_sparkPositions;//local

    //b2Fixture* m_sparkFixture;
    vector<bool> m_sparkOn;

    vector<b2Fixture*> m_valveFixtures;
    vector<b2Fixture*> m_blockerFixtures;
    vector<bool> m_valvesOpen;

    vector< pair<b2Fixture*, b2Fixture*> > m_ignitionFixturePairs;

    vector<b2Fixture*> m_catchers;

    float m_rotationSinceLastFuel;
    float m_fuelRate;

    set<b2Body*> m_fuelParticles;
    vector<b2Body*> m_particlesToRecycle;
    vector<b2Body*> m_queuedParticles;

    b2Body* m_torqueGuageBody;
    b2FrictionJoint* m_torqueGuageFrictionJoint;
    b2MotorJoint* m_torqueGuageMotorJoint;

    float m_currentTorque; //average of last five seconds
    float m_currentRpm; //average of last five seconds
    float m_currentFriction; //average of last five seconds

    float m_torqueForGraph; //average of last 10 steps

    float m_torqueMeasurements[600];//ten seconds

    b2Fixture* m_waterFixture;
    b2Fixture* m_boathullFixture;
    std::set<fixturePair> m_buoyancyFixturePairs;

    //b2PrismaticJoint* m_brakePistonJoint;
    //b2PrismaticJoint* m_brakePadJoint;

    bool m_spinCrank;

    bool m_runningDyno;
};

#endif
